// Description: This is a simple benchmark based on VPIC's I/O interface
//		Each process writes a specified number of particles into 
//		a h5part output file
// Author:	Suren Byna <SByna@lbl.gov>
//		Lawrence Berkeley National Laboratory, Berkeley, CA
// 


#include "H5Part.h"
#include "H5Block.h"
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include "./timer.h"

struct timeval start_time[3];
float elapse[3];

int num_steps = 5;
int persist_rate = 1;

H5PartFile* file;
int numparticles = 8388608;	// 8  meg particles
float *x, *y, *z;
float *px, *py, *pz;
h5part_int32_t *id1, *id2;
float *rx, *ry, *rz;
float *rpx, *rpy, *rpz;
h5part_int32_t *rid1, *rid2;
int x_dim = 64;
int y_dim = 64; 
int z_dim = 64;
char* file_name = NULL;
int num_procs = 0, split_size = 0;
int num_writers = 0;
int quit_on_corr = 0;

void corrupt_execute(char *arg)
{
    pid_t  pid;
    int    status;
    char *argv[] = {"ssh","lola-12","/scratch/iod/bin/iod_corrupt_tool",
                    "arg",NULL};

    if (fork()) {
        wait(&status);
    } else {
        execvp("/usr/bin/ssh", argv);
    }
}

// Uniform random number
inline double uniform_random_number() 
{
	return (((double)rand())/((double)(RAND_MAX)));
}

// Initialize particle data
void init_particles ()
{
	int i;
	for (i=0; i<numparticles; i++) 
	{
            id1[i] = i;
            id2[i] = i*2;
            x[i] = i*4;
            y[i] = i*5;
            z[i] = i+5;
            px[i] = i*12;
            py[i] = i+1;
            pz[i] = i+10;
            /*
            x[i] = uniform_random_number()*x_dim;
            y[i] = uniform_random_number()*y_dim;
            z[i] = ((double)id1[i]/numparticles)*z_dim;
            px[i] = uniform_random_number()*x_dim;
            py[i] = uniform_random_number()*y_dim;
            pz[i] = ((double)id2[i]/numparticles)*z_dim;
            */
	}
}

int create_named_datatype(int step, hid_t trans_id) {
    hid_t dtid;
    char dtname[5];

    dtid = H5Tcopy(H5T_STD_I32LE);
    sprintf(dtname, "DT%d", step);

    assert(H5Tcommit_ff(file->file, dtname, dtid, H5P_DEFAULT, H5P_DEFAULT, 
                        H5P_DEFAULT, trans_id, H5_EVENT_STACK_NULL) == 0);
    assert(H5Tclose_ff(dtid, H5_EVENT_STACK_NULL) == 0);

    return 0;
}

int check_named_datatype(int step, hid_t rc_id) {
    hid_t dtid;
    char dtname[5];

    sprintf(dtname, "DT%d", step);

    dtid = H5Topen_ff(file->file, dtname, H5P_DEFAULT, rc_id, H5_EVENT_STACK_NULL);
    if(dtid < 0) {
        fprintf (stderr, "Blob corruption detected when opening Datatype %s\n", dtname);
        if(quit_on_corr)
            exit(1);
        return -1;
    }

    assert(H5Tclose_ff(dtid, H5_EVENT_STACK_NULL) == 0);

    return 0;
}

// Create H5Part file and write data
void create_and_write_synthetic_h5part_data(int step, int rank, hid_t rc_id, hid_t trans_id) 
{
    h5part_int64_t ret;

    ret = H5PartSetStep_ff(file, step, rc_id, trans_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set timestep %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Step %d is set \n", step);

    ret = H5PartSetNumParticles(file, numparticles/split_size);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set particales step %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Num particles is set \n");

    ret = H5PartWriteDataInt32_ff(file,"id1",&id1[rank*numparticles/split_size], rc_id, trans_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set timestep %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Wrote variable 1 \n");
    ret = H5PartWriteDataInt32_ff(file,"id2",&id2[rank*numparticles/split_size], rc_id, trans_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set timestep %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Wrote variable 2 \n");

    ret = H5PartWriteDataFloat32_ff(file,"x",&x[rank*numparticles/split_size], rc_id, trans_id); 
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set timestep %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Wrote variable 3 \n");
    ret = H5PartWriteDataFloat32_ff(file,"y",&y[rank*numparticles/split_size], rc_id, trans_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set timestep %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Wrote variable 4 \n");
    ret = H5PartWriteDataFloat32_ff(file,"z",&z[rank*numparticles/split_size], rc_id, trans_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set timestep %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Wrote variable 5 \n");

    ret = H5PartWriteDataFloat32_ff(file,"px",&px[rank*numparticles/split_size], rc_id, trans_id); 
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set timestep %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Wrote variable 6 \n");
    ret = H5PartWriteDataFloat32_ff(file,"py",&py[rank*numparticles/split_size], rc_id, trans_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set timestep %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Wrote variable 7 \n");
    ret = H5PartWriteDataFloat32_ff(file,"pz",&pz[rank*numparticles/split_size], rc_id, trans_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set timestep %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Wrote variable 8 \n");
}

int read_and_verify_synthetic_h5part_data(int step, int rank, hid_t rc_id) 
{
    int start, end, i;
    size_t num_errs = 0;
    h5part_int64_t ret;

    memset(rx, 0, numparticles*sizeof(double));
    memset(ry, 0, numparticles*sizeof(double));
    memset(rz, 0, numparticles*sizeof(double));
    memset(rid1, 0, numparticles*sizeof(h5part_int32_t));
    memset(rid2, 0, numparticles*sizeof(h5part_int32_t));
    memset(rpx, 0, numparticles*sizeof(double));
    memset(rpy, 0, numparticles*sizeof(double));
    memset(rpz, 0, numparticles*sizeof(double));

    ret = H5PartSetStep_ff(file, step, rc_id, 0);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set timestep %d\n", step);
        exit(1);
    }
    else if(ret != H5PART_SUCCESS) {
        fprintf (stderr, "Failed to set Step %d\n", step);
        return 1;
    }

    if (rank == 0) 
        fprintf (stderr, "Step %d is set \n", step);

    ret = H5PartSetNumParticles(file, numparticles/split_size);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to set particales step %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Num particles is set \n");

    start = floor(rank*numparticles/split_size);
    end = floor((rank+1)*numparticles/split_size);

    ret = H5PartReadDataInt32_ff(file,"id1",&rid1[start], rc_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to read data step %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Read variable 1 \n");
    ret = H5PartReadDataInt32_ff(file,"id2",&rid2[start], rc_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to read data step %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Read variable 2 \n");

    ret = H5PartReadDataFloat32_ff(file,"x",&rx[start], rc_id); 
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to read data step %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Read variable 3 \n");
    ret = H5PartReadDataFloat32_ff(file,"y",&ry[start], rc_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to read data step %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Read variable 4 \n");
    ret = H5PartReadDataFloat32_ff(file,"z",&rz[start], rc_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to read data step %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Read variable 5 \n");

    ret = H5PartReadDataFloat32_ff(file,"px",&rpx[start], rc_id); 
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to read data step %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Read variable 6 \n");
    ret = H5PartReadDataFloat32_ff(file,"py",&rpy[start], rc_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to read data step %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Read variable 7 \n");
    ret = H5PartReadDataFloat32_ff(file,"pz",&rpz[start], rc_id);
    if(ret != H5PART_SUCCESS && quit_on_corr) {
        fprintf (stderr, "Failed to read data step %d\n", step);
        exit(1);
    }
    if (rank == 0) fprintf (stderr, "Read variable 8 \n");

    for (i=start; i<end; i++) {

        if(x[i] != rx[i]) {fprintf (stderr, "Corruption variable x at %d (%d vs %d)\n",
                                    i, (int)x[i], (int)rx[i]); num_errs ++;}
        if(y[i] != ry[i]) {fprintf (stderr, "Corruption variable y \n"); num_errs ++;}
        if(z[i] != rz[i]) {fprintf (stderr, "Corruption variable z \n"); num_errs ++;}
        if(id1[i] != rid1[i]) {fprintf (stderr, "Corruption variable id1 \n"); num_errs ++;}
        if(id2[i] != rid2[i]) {fprintf (stderr, "Corruption variable id2 \n"); num_errs ++;}
        if(px[i] != rpx[i]) {fprintf (stderr, "Corruption variable px \n"); num_errs ++;}
        if(py[i] != rpy[i]) {fprintf (stderr, "Corruption variable py \n"); num_errs ++;}
        if(pz[i] != rpz[i]) {fprintf (stderr, "Corruption variable pz \n"); num_errs ++;}

        if(num_errs != 0) {
            if(quit_on_corr)
                exit(1);
            return(1);
        }
    }

    return 0;
}

static int
parse_options(int argc, char** argv) 
{
    int c;
    extern char *optarg;

    while ((c = getopt (argc, argv, "p:s:r:f:w:")) != -1) {
        switch (c) {
        case 'p':
            numparticles = atoi(optarg)*1024*1024;
            break;
        case 's':
            num_steps = atoi(optarg);
            break;
        case 'r':
            persist_rate = atoi(optarg);
            break;
        case 'f':
            file_name = malloc(100);//strdup(optarg);
            sprintf(file_name, "%s.%s", getenv("USER"), optarg);
            printf("File_Name :%s\n", file_name);
            break;
        case 'w':
            num_writers = atoi(optarg);
            break;
        default:
            abort ();
        }
    }
    return( 0 );
}

int main (int argc, char* argv[]) 
{
        hid_t trans_id, rc_id;
	int my_rank, provided, split_rank;
        int i;
        uint64_t c_version;
        h5part_int64_t status;
        MPI_Comm scomm;
        int color;

        MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &provided);
        if(MPI_THREAD_MULTIPLE != provided) {
            fprintf(stderr, "MPI does not have MPI_THREAD_MULTIPLE support\n");
            exit(1);
        }

        /* Call EFF_init to initialize the EFF stack. */
        EFF_init(MPI_COMM_WORLD, MPI_INFO_NULL);

	MPI_Comm_rank (MPI_COMM_WORLD, &my_rank);
	MPI_Comm_size (MPI_COMM_WORLD, &num_procs);

        num_writers = num_procs/2;

        color = my_rank % 2;
        MPI_Comm_split( MPI_COMM_WORLD, color, my_rank, &scomm );
	MPI_Comm_size (scomm, &split_size);
	MPI_Comm_rank (scomm, &split_rank);

        if (parse_options(argc, argv) != 0 ) {
            exit( 1 );
        }

        if(NULL == file_name) {
            fprintf (stderr, "Need to pass file_name with -f\n");
            exit(1);
        }

        if(num_procs % num_writers != 0) {
            fprintf (stderr, "total procs must be divisible by number of writers\n");
            exit(1);
        }

	if (my_rank == 0) 
            fprintf (stderr, "Number of paritcles: %ld (%ld per rank).\n", 
                     numparticles, numparticles/split_size);

	x=(float*)malloc(numparticles*sizeof(double));
	y=(float*)malloc(numparticles*sizeof(double));
	z=(float*)malloc(numparticles*sizeof(double));

	px=(float*)malloc(numparticles*sizeof(double));
	py=(float*)malloc(numparticles*sizeof(double));
	pz=(float*)malloc(numparticles*sizeof(double));

	id1=(h5part_int32_t*)malloc(numparticles*sizeof(h5part_int32_t));
	id2=(h5part_int32_t*)malloc(numparticles*sizeof(h5part_int32_t));

        if(color == 1) {
            rx=(float*)malloc(numparticles*sizeof(double));
            ry=(float*)malloc(numparticles*sizeof(double));
            rz=(float*)malloc(numparticles*sizeof(double));

            rpx=(float*)malloc(numparticles*sizeof(double));
            rpy=(float*)malloc(numparticles*sizeof(double));
            rpz=(float*)malloc(numparticles*sizeof(double));

            rid1=(h5part_int32_t*)malloc(numparticles*sizeof(h5part_int32_t));
            rid2=(h5part_int32_t*)malloc(numparticles*sizeof(h5part_int32_t));
        }

	init_particles ();

	if (my_rank == 0)
            fprintf (stderr, "Finished initializeing particles \n");

	MPI_Barrier (MPI_COMM_WORLD);

        timer_on (0);
        file = H5PartOpenFileParallel_ff (file_name, H5PART_WRITE, scomm, &rc_id);
        if(NULL == file) {
            fprintf (stderr, "Failed to open H5Part file\n");
            exit(1);
        }

        H5RCget_version(rc_id, &c_version);

        MPI_Barrier(MPI_COMM_WORLD);

	if (my_rank == 0)
	{
            fprintf (stderr, "Opened H5Part file... \n");
            fprintf(stderr, "Container version at %d \n", (int)c_version);
	}

        if(c_version == 1) {
            if(0 == color) {
                trans_id = H5PartTRstart(file, rc_id, (uint64_t)2);
                assert(trans_id > 0);

                if(my_rank == 0)
                    H5PartWriteFileAttribString_ff(file, "Origin", "Tested by Mohamad", 
                                                   rc_id, trans_id);

                H5PartTRfinish(trans_id);
            }

            MPI_Barrier (MPI_COMM_WORLD);

            timer_on (1);
            if (my_rank == 0) fprintf (stderr, "Before writing particles \n");

            for(i=0 ; i<num_steps; i++) {
                hid_t temp_rc;

                if (my_rank == 0) 
                    fprintf (stderr, "Step %d (transaction %d):\n", i, i+3);

                if(0 == color) {
                    trans_id = H5PartTRstart(file, rc_id, (uint64_t)i+3);
                    assert(trans_id > 0);

                    create_and_write_synthetic_h5part_data(i, split_rank, rc_id, trans_id);

                    if(0 == my_rank)
                        create_named_datatype(i, trans_id);

                    H5PartTRfinish(trans_id);
                }

                MPI_Barrier (MPI_COMM_WORLD);

                if(my_rank == 0 && (i+1)%persist_rate == 0) {
                    uint64_t rc_version = (uint64_t)i+3;

                    fprintf (stderr, "Persist After Step %d Transaction %d\n", i, i+3);
                    temp_rc = H5PartRCacquire(file, (uint64_t)i+3);
                    assert(temp_rc > 0);

                    MPI_Send(&rc_version, 1, MPI_UINT64_T, 1, 123, MPI_COMM_WORLD);

                    assert(H5RCpersist(temp_rc, H5_EVENT_STACK_NULL) == 0);
                }

                if(color == 1 && (i+1)%persist_rate == 0) {
                    uint64_t rc_version;
                    int end_step, j;

                    if(my_rank == 1) {
                        MPI_Recv(&rc_version, 1, MPI_UINT64_T, 0, 123, 
                                 MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                    }

                    MPI_Bcast(&rc_version, 1, MPI_UINT64_T, 0, scomm);
                    temp_rc = H5RCcreate(file->file, rc_version);

                    end_step = (int)rc_version - 2;

                    for(j=end_step-persist_rate ; j<end_step; j++) {
                        if (my_rank == 1) 
                            fprintf (stderr, "Step %d\n", j);

                        if(read_and_verify_synthetic_h5part_data(j, split_rank, temp_rc) != 0) {
                            fprintf (stderr, "Step Verification Failed. Data Corruption was detected!\n");
                        }

                        if(check_named_datatype(j, temp_rc) != 0) {
                            fprintf (stderr, "Step Verification Failed. Data Corruption was detected!\n");
                        }
                    }
                    H5RCclose(temp_rc);
                }
                MPI_Barrier (MPI_COMM_WORLD);
                if(my_rank == 0 && (i+1)%persist_rate == 0) {
                    //assert(H5RCpersist(temp_rc, H5_EVENT_STACK_NULL) == 0);
                    assert(H5PartRCrelease(temp_rc) == 0);
                }
            }
        }
        else if((int)c_version >= num_steps+2) {
            fprintf(stderr, "No more steps to write!\n");
        }
        else {
            int step_start;
            hid_t temp_rc;

            step_start = (int)c_version - 2;
            for(i=step_start ; i<num_steps; i++) {
                if (my_rank == 0) 
                    fprintf (stderr, "Step %d (transaction %d): \n", i, i+3);

                if(0 == color) {
                    trans_id = H5PartTRstart(file, rc_id, (uint64_t)i+3);
                    assert(trans_id > 0);

                    create_and_write_synthetic_h5part_data(i, split_rank, rc_id, trans_id);

                    if(0 == my_rank)
                        create_named_datatype(i, trans_id);

                    H5PartTRfinish(trans_id);
                }

                MPI_Barrier (MPI_COMM_WORLD);

                if(my_rank == 0 && (i+1)%persist_rate == 0) {
                    uint64_t rc_version = (uint64_t)i+3;

                    fprintf (stderr, "Persist After Step %d Transaction %d\n", i, i+3);
                    temp_rc = H5PartRCacquire(file, (uint64_t)i+3);
                    assert(temp_rc > 0);

                    MPI_Send(&rc_version, 1, MPI_UINT64_T, 1, 123, MPI_COMM_WORLD);
                    assert(H5RCpersist(temp_rc, H5_EVENT_STACK_NULL) == 0);
                }

                if(color == 1 && (i+1)%persist_rate == 0) {
                    uint64_t rc_version;
                    int end_step, j;

                    if(my_rank == 1) {
                        MPI_Recv(&rc_version, 1, MPI_UINT64_T, 0, 123, 
                                 MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                    }

                    MPI_Bcast(&rc_version, 1, MPI_UINT64_T, 0, scomm);
                    temp_rc = H5RCcreate(file->file, rc_version);

                    end_step = (int)rc_version - 2;

                    for(j=end_step-persist_rate ; j<end_step; j++) {
                        if (my_rank == 1) 
                            fprintf (stderr, "Step %d\n", j);

                        if(read_and_verify_synthetic_h5part_data(j, split_rank, temp_rc) != 0) {
                            fprintf (stderr, "Step Verification Failed. Data Corruption was detected!\n");
                        }

                        if(check_named_datatype(j, temp_rc) != 0) {
                            fprintf (stderr, "Step Verification Failed. Data Corruption was detected!\n");
                        }
                    }
                    H5RCclose(temp_rc);
                }
                MPI_Barrier (MPI_COMM_WORLD);
                if(my_rank == 0 && (i+1)%persist_rate == 0) {
                    assert(H5PartRCrelease(temp_rc) == 0);
                }
            }
        }

	MPI_Barrier (MPI_COMM_WORLD);
        H5PartRCrelease(rc_id);
	MPI_Barrier (MPI_COMM_WORLD);
	timer_off (1);

	if (my_rank == 0) 
            fprintf (stderr, "After writing particles \n");

        H5PartCloseFile(file);

	MPI_Barrier (MPI_COMM_WORLD);

	if (my_rank == 0) 
            fprintf (stderr, "After closing H5Part file \n");

	free(x); free(y); free(z);
	free(px); free(py); free(pz);
	free(id1);
	free(id2);
	free(rx); free(ry); free(rz);
	free(rpx); free(rpy); free(rpz);
	free(rid1);
	free(rid2);

        EFF_finalize();

	timer_off (0);

	if (my_rank == 0)
	{
		printf ("\nTiming results\n");
		timer_msg (1, "just writing data");
		timer_msg (0, "opening, writing, closing file");
		printf ("\n");
	}

        free(file_name);

	MPI_Finalize();

	return 0;
}
